/*
 * Copyright (c) 2005-2010 KOM � Multimedia Communications Lab
 *
 * This file is part of Simonstrator.KOM.
 * 
 * Simonstrator.KOM is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 * 
 * PeerfactSim.KOM is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with PeerfactSim.KOM.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package de.tudarmstadt.maki.simonstrator.api.component.relevance.vehicular.impl;

import java.util.List;

import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.TemporalDependencyMatrix;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.measurement.MeasurementDistributionTypeContainer;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.properties.vectoral.VectoralProperty;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.information.PointInformation;

public class VectoralQoIBasedImpactFunction<T extends PointInformation> extends AbstractQoIBasedImpactFunction<T> {

	private static final double MIN_STEP = 0.0001;

	public VectoralQoIBasedImpactFunction(List<T> pInformation, double pAccuracy, int pMaxCacheSize,
			Object pLastDecision) {
		super(pInformation, pAccuracy, pMaxCacheSize, pLastDecision);
	}

	@Override
	protected double determineB(Object pTemplate, Object pPossibleValue, double pChangeRate, double pRate,
			double pErrorProbability, double pNumberOfMessages, double pCostWrongKeep, double pCostWrongChange) {
		if (!(pTemplate instanceof VectoralProperty)) {
			throw new AssertionError("Vectoral property required as input for pTemplate, but got "
					+ pTemplate.getClass().getSimpleName());
		}
		return determineBWithCast((VectoralProperty) pTemplate, (int) pPossibleValue, pChangeRate, pRate,
				pErrorProbability, pNumberOfMessages, pCostWrongKeep, pCostWrongChange);
	}

	public double determineBWithCast(VectoralProperty pTemplate, int pPossibleValue, double change, double rate,
			double errorProbability, double pNumberOfMessages, double costSlow, double costFast) {
		if (errorProbability == 0 || errorProbability == 1 || errorProbability == 0.5) {
			return Double.NaN;
		}

		if (_lastDecision != null) {
			if (pPossibleValue == ((VectoralProperty) _lastDecision).getMostProbableIndex()) {
				return Double.NaN;
			}
		} else {
			if (pPossibleValue == pTemplate.getDefaultIndex()) {
				return Double.NaN;
			}
		}

		double b;
		double p_c = change;

		int optimalAmount = getOptimalMessageAmountForSwitch(p_c, errorProbability, costSlow, costFast);

		if (maxCacheSize <= optimalAmount) {
			return Double.POSITIVE_INFINITY;
		}

		if ((int) pNumberOfMessages <= optimalAmount) {
			return Double.POSITIVE_INFINITY;
		}

		if (optimalAmount == 1) {
			return Double.NEGATIVE_INFINITY;
		}

		boolean first = true;

		double step = 10;

		if (errorProbability < 0.5) {
			b = -1 * step;
		} else {
			b = 1 * step;
		}

		do {
			VectoralProperty valueAfterN = calculateMostProbable(pTemplate, pPossibleValue, optimalAmount, rate,
					errorProbability, pNumberOfMessages, b);
			double probableValueAfterN = Double.NaN;
			if (valueAfterN != null) {
				probableValueAfterN = valueAfterN.getMostProbableIndex();
			}
			VectoralProperty valueBeforeN = calculateMostProbable(pTemplate, pPossibleValue, optimalAmount - 1, rate,
					errorProbability, pNumberOfMessages, b);
			double probableValueBeforeN = Double.NaN;
			if (valueBeforeN != null) {
				probableValueBeforeN = valueBeforeN.getMostProbableIndex();
			}

			if (probableValueAfterN == pPossibleValue && probableValueAfterN != probableValueBeforeN) {
				return b;
			}

			if (!Double.isNaN(probableValueBeforeN) && !Double.isNaN(probableValueBeforeN) && step >= MIN_STEP) {
				if (probableValueAfterN != pPossibleValue) {
					if (b < 0) {
						b -= step;
						if (!first) {
							step *= 0.5;
						}
					} else {
						b -= step;
						step *= 0.5;
						first = false;
					}
				} else if (probableValueAfterN == pPossibleValue && probableValueAfterN == probableValueBeforeN) {
					if (b > 0) {
						b += step;
						if (!first) {
							step *= 0.5;
						}
					} else {
						b += step;
						step *= 0.5;
						first = false;
					}
				} else {
					return Double.NaN;
				}
			} else {
				return Double.NaN;
			}
		} while (true);
	}

	public VectoralProperty calculateMostProbable(VectoralProperty pTemplate, int pNewValue, int optimalMessageAmount,
			double rate, double errorProbability, double pNumberOfMessages, double b) {
		VectoralProperty currentProperty = null;

		for (int a = optimalMessageAmount + 1; a <= Math.min(pNumberOfMessages, maxCacheSize); a++) {
			VectoralProperty jamProperty = pTemplate.clone();

			if (_lastDecision != null) {
				jamProperty.set(((VectoralProperty) _lastDecision).getMostProbableIndex(), accuracy,
						MeasurementDistributionTypeContainer.getDistribution(pTemplate.getClass()));
			} else {
				jamProperty.set(pTemplate.getDefaultIndex(), accuracy,
						MeasurementDistributionTypeContainer.getDistribution(pTemplate.getClass()));
			}
			TemporalDependencyMatrix temporalDependencyMatrix = jamProperty.getDependencyMatrix();
			temporalDependencyMatrix = temporalDependencyMatrix.age((long) (a * rate));
			double impact = calculateImpact(errorProbability, pNumberOfMessages, pNumberOfMessages - a, b);

			if (Double.isNaN(impact)) {
				return null;
			}

			temporalDependencyMatrix = modifyDependencyMatrix(temporalDependencyMatrix, impact);
			jamProperty = (VectoralProperty) jamProperty.age(1, temporalDependencyMatrix);

			if (currentProperty != null) {
				currentProperty = (VectoralProperty) currentProperty.combine(jamProperty);
			} else {
				currentProperty = jamProperty;
			}
		}

		for (int a = 0; a <= optimalMessageAmount; a++) {
			VectoralProperty jamProperty = pTemplate.clone();
			jamProperty.setGaussianWithAccuracy(pNewValue, accuracy);

			TemporalDependencyMatrix temporalDependencyMatrix = jamProperty.getDependencyMatrix();
			temporalDependencyMatrix = temporalDependencyMatrix.age((long) (a * rate));
			double impact = calculateImpact(errorProbability, pNumberOfMessages, pNumberOfMessages - a, b);

			if (Double.isNaN(impact)) {
				return null;
			}

			temporalDependencyMatrix = modifyDependencyMatrix(temporalDependencyMatrix, impact);
			jamProperty = (VectoralProperty) jamProperty.age(1, temporalDependencyMatrix);

			if (currentProperty != null) {
				currentProperty = (VectoralProperty) currentProperty.combine(jamProperty);
			} else {
				currentProperty = jamProperty;
			}
		}

		return currentProperty;
	}

	private TemporalDependencyMatrix modifyDependencyMatrix(TemporalDependencyMatrix pDependencyMatrix,
			double pImpact) {
		TemporalDependencyMatrix result = pDependencyMatrix.clone();

		double[][] dependencies = result.getDependencies();
		for (int i = 0; i < dependencies.length; i++) {
			double finalPercentages = 1.0 / dependencies[i].length;

			for (int j = 0; j < dependencies[i].length; j++) {
				dependencies[i][j] = finalPercentages
						+ (pDependencyMatrix.getDependencies()[i][j] - finalPercentages) * pImpact;
			}
		}

		return result;
	}
}
