/*
 * Copyright (c) 2005-2010 KOM – Multimedia Communications Lab
 *
 * This file is part of Simonstrator.KOM.
 * 
 * Simonstrator.KOM is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 * 
 * PeerfactSim.KOM is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with PeerfactSim.KOM.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.costs;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import de.tudarmstadt.maki.simonstrator.api.component.pubsub.Notification;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.properties.BumpProperty;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.properties.FogProperty;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.properties.HazardProperty;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.properties.JamProperty;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.properties.LocationBasedEnvironmentProperty;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.properties.RainProperty;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.properties.RoadProperty;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.properties.TrafficSignProperty;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.information.EnvironmentInformation;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.information.RoadInformation;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.roadnetwork.RoadNetwork;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.roadnetwork.RoadNetworkEdge;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.roadnetwork.RoadNetworkRoute;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.roadnetwork.paths.VehiclePathTracker;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.roadnetwork.paths.VehiclePathTrackerFactory;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.roadnetwork.routing.DijkstraAlgorithm;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.roadnetwork.routing.RoutingAlgorithm;

public class DefaultVehicularPropertyCostEstimator implements VehicularPropertyCostEstimator {

	private static final int MAX_DETOUR = 3;

	private RoutingAlgorithm routing = new DijkstraAlgorithm();

	@Override
	public double calculateCostsIfKnown(Class<? extends RoadProperty> pProperty) {
		if (pProperty.equals(JamProperty.class)) {
			return 3 * Math.pow(10, 2);
		} else if (pProperty.equals(RainProperty.class)) {
			return 1;
		} else if (pProperty.equals(FogProperty.class)) {
			return 1;
		} else if (pProperty.equals(HazardProperty.class)) {
			return Math.pow(10, 3);
		} else if (pProperty.equals(TrafficSignProperty.class)) {
			return 10;
		}
		return 10;
	}

	@Override
	public double calculateCostsIfKnown(Notification pNotification) {
		EnvironmentInformation<? extends LocationBasedEnvironmentProperty> environmentInformation = EnvironmentInformation
				.createFromNotification(pNotification);

		if (environmentInformation instanceof RoadInformation) {
			return calculateCostsIfKnown(((RoadInformation) environmentInformation).getValue());
		}

		throw new AssertionError("Cost calculation only valid for road information!");
	}

	@Override
	public List<DecisionPoint> getDecisionPoints(Notification pNotification) {
		EnvironmentInformation<? extends LocationBasedEnvironmentProperty> environmentInformation = EnvironmentInformation
				.createFromNotification(pNotification);

		if (environmentInformation instanceof RoadInformation) {
			RoadInformation roadInformation = (RoadInformation) environmentInformation;

			List<DecisionPoint> decisionPoints = new ArrayList<>();

			decisionPoints.add(new VehicularDecisionPoint(roadInformation.getEdge(), roadInformation.getLocation(),
					calculateCostsIfKnown(roadInformation.getValue())));

			if (roadInformation.getValue().isDetourable()) {
				Set<RoadNetworkEdge> edges = new HashSet<>();
				edges.add(roadInformation.getEdge());
				for (int i = 0; i < MAX_DETOUR; i++) {
					Set<RoadNetworkEdge> newEdges = new HashSet<>();
					for (RoadNetworkEdge roadNetworkEdge : edges) {
						newEdges.addAll(roadNetworkEdge.getIncomingEdges());
					}
					edges.addAll(newEdges);
				}

				for (RoadNetworkEdge roadNetworkEdge : edges) {
					decisionPoints.add(new VehicularDetouringDecisionPoint(roadNetworkEdge, roadInformation.getEdge(),
							roadInformation.getLocation(), calculateCostsIfKnown(roadInformation.getValue())));
				}
			}

			return decisionPoints;
		}

		throw new AssertionError("Cost calculation only valid for road information!");
	}

	@Override
	public double calculateCostsIfKnown(RoadProperty pProperty, RoadNetworkEdge pDecisionPoint,
			VehicleCostPreference pVehicle, RoadNetworkRoute pPath) {
		VehiclePathTracker tracker = VehiclePathTrackerFactory.getVehiclePathTracker();
		if (pPath.containsEdge(pDecisionPoint) && pPath.containsEdge(pProperty.getEdge())) {
			if (pPath.getStart().equals(pProperty.getEdge())) {
				return calculateCostsIfKnown(pProperty);
			}
			if (pProperty.isDetourable()) {
				// Find other route...
				List<RoadNetworkEdge> edgesToAvoid = new ArrayList<>();
				edgesToAvoid.add(pProperty.getEdge());
				RoadNetworkRoute route = routing.findRoute(RoadNetwork.CURRENT_ROAD_NETWORK, pPath.getStart(),
						pPath.getDestination(), Collections.emptyList(), edgesToAvoid);

				if (route != null) {
					double newRouteTravelTime = route.getTravelTime();
					double routeTravelTime = pPath.getTravelTime();

					boolean found = false;
					for (int i = 0; i < pPath.getRoute().size() - 1; i++) {
						if (pPath.getRoute().get(i).equals(pProperty.getEdge())) {
							routeTravelTime -= tracker.getTravelTime(pPath.getRoute().get(i),
									pPath.getRoute().get(i + 1));
							found = true;
						}
					}
					if (found) {
						routeTravelTime += (pProperty.getEdge().getLength() / pProperty.getEdge().getMaxSpeed());
					}

					if (routeTravelTime > newRouteTravelTime) {
						return calculateCostsIfKnown(pProperty) + (routeTravelTime - newRouteTravelTime) * 10;
					}
				}
			}
		}

		return calculateCostsIfKnown(pProperty);
	}

	@Override
	public double calculateCostsIfUnknown(Notification pNotification) {
		EnvironmentInformation<? extends LocationBasedEnvironmentProperty> environmentInformation = EnvironmentInformation
				.createFromNotification(pNotification);

		if (environmentInformation instanceof RoadInformation) {
			return calculateCostsIfUnknown(((RoadInformation) environmentInformation).getValue());
		}

		throw new AssertionError("Cost calculation only valid for road information!");
	}

	@Override
	public double calculateCostsIfUnknown(Class<? extends RoadProperty> pProperty) {
		if (pProperty.equals(JamProperty.class)) {
			return Math.pow(10, 4);
		} else if (pProperty.equals(BumpProperty.class)) {
			return Math.pow(10, -2);
		} else if (pProperty.equals(HazardProperty.class)) {
			return Math.pow(10, 6);
		}
		return 1;
	}

	@Override
	public double calculateCostsIfUnknown(RoadProperty pProperty, RoadNetworkEdge pDecisionPoint,
			VehicleCostPreference pVehicle, RoadNetworkRoute pPath) {
		return 0;
	}

}
